from typing import Any

from .async_http_client_contract import AsyncHTTPClientContract
from .clients.aiohttp_client import AioHttpClient
from .types.async_http_client_types import (
    AsyncHttpClientInitArgs,
    GetArgs,
    UploadFileArgs,
)


class AsyncHTTPClient(AsyncHTTPClientContract):
    CLIENTS = {"aiohttp_client"}

    def __init__(self, args: AsyncHttpClientInitArgs) -> None:
        if args.client_name not in AsyncHTTPClient.CLIENTS:
            msg = f"Unsupported client '{args.client_name}'"
            raise KeyError(msg)
        self.client_name = args.client_name

        if args.client_name == "aiohttp_client":
            self.client_obj = AioHttpClient(args)

    async def __aenter__(self):
        return await self.client_obj.__aenter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client_obj.__aexit__(exc_type, exc_val, exc_tb)

    async def get(self, args: GetArgs) -> dict[str, Any]:
        return await self.client_obj.get(args)

    async def upload_file(self, args: UploadFileArgs) -> dict[str, Any]:
        return await self.client_obj.upload_file(args)



from typing import Any, Dict

from .controllers.types.update_thies_data_types import UpdateThiesDataControllerInput
from .controllers.update_thies_data import UpdateThiesDataController
from rcer_iot_client_pkg.general_types.api.update_thies_data_types import EpiiAPIConfig


class EpiiAPI:
    """
    EpiiAPI is a service class that provides methods to interact with Patagonia Center system.
    """

    async def update_thies_data(self, config: EpiiAPIConfig) -> Dict[str, Any]:
        """
        This method establishes a connection to an FTP server using the provided
        credentials and updates data related to THIES Data Logger.
        Args:
            config (EpiiAPIConfig): configuration class for FTP Server and Microsoft SharePoint credentials.
        Returns:
            response (dict): A dictionary representation of the API response.
        """
        controller = UpdateThiesDataController(UpdateThiesDataControllerInput(config))
        response = await controller.execute()
        return response.__dict__



"""Export defined type classes."""

from .async_http_client import AsyncHTTPClient
from .types.async_http_client_types import (
    AsyncHttpClientInitArgs,
    GetArgs,
    UploadFileArgs,
)

__all__ = ["AsyncHTTPClient", "AsyncHttpClientInitArgs", "GetArgs", "UploadFileArgs"]



class ThiesConnectionError(Exception):
    """Raised when unable to connect to the THIES FTP Server"""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return "Unable to connect to THIES FTP Server. " + self.reason.__str__()


class ThiesFetchingError(Exception):
    """Raised when no files are found to upload to the server."""

    def __init__(self, *args, reason):
        super().__init__(*args, reason)
        self.reason = reason

    def __str__(self):
        return (
            "An error ocurred while retrieving files from THIES FTP Server. "
            + self.reason.__str__()
        )


class SharePointFetchingError(Exception):
    """Raised when there is an error fetching file names from the RCER cloud."""

    def __str__(self):
        return "An error occurred while retrieving file names from the RCER cloud"



from dotenv import load_dotenv

import rcer_iot_client_pkg.services.epii.constants.update_thies_data_constants as c
from rcer_iot_client_pkg.general_types.error_types.api.update_thies_data_error_types import (
    SharePointFetchingError,
    ThiesConnectionError,
    ThiesFetchingError,
)
from rcer_iot_client_pkg.general_types.error_types.common import (
    EmptyDataError,
    FtpClientError,
    SharepointClientError,
)
from rcer_iot_client_pkg.libs.ftp_client import (
    FTPClient,
    FtpClientInitArgs,
    FtpListFilesArgs,
    FtpReadFileArgs,
)
from rcer_iot_client_pkg.libs.sharepoint_client import (
    SharepointClient,
    SharepointClientInitArgs,
    SpListFilesArgs,
)
from rcer_iot_client_pkg.services.epii.use_cases.types import (
    UpdateThiesDataUseCaseInput,
    FtpClientConfig,
    SharepointConfig,
)
from rcer_iot_client_pkg.services.epii.utils import (
    parse_execute_response,
)

load_dotenv()


class UpdateThiesDataUseCase:
    def __init__(self, input: UpdateThiesDataUseCaseInput):
        self.sharepoint_client = self._initialize_sharepoint_client(
            input.sharepoint_config
        )
        self.thies_ftp_client = self._initialize_thies_ftp_client(input.ftp_config)
        self.uploading = set()

    def _initialize_sharepoint_client(
        self, config: SharepointConfig
    ) -> SharepointClient:
        """Initialize the HTTP client."""
        try:
            return SharepointClient(
                SharepointClientInitArgs(config, client_name="sharepoint_rest_api")
            )
        except ConnectionError as error:
            raise SharepointClientError(error)

    def _initialize_thies_ftp_client(self, config: FtpClientConfig) -> FTPClient:
        """Initialize the FTP client."""
        try:
            return FTPClient(FtpClientInitArgs(config, client_name="aioftp_client"))
        except RuntimeError as error:
            raise FtpClientError(error)

    async def fetch_cloud_file_names(self) -> set[str]:
        """Fetch file names from the RCER cloud."""

        try:
            cloud_files = set()
            async with self.sharepoint_client:
                for folder in c.SHAREPOINT_THIES_FOLDERS:
                    args = SpListFilesArgs(
                        folder_relative_url=f"{c.SHAREPOINT_BASE_URL}/{folder}"
                    )
                    response = await self.sharepoint_client.list_files(args)
                    cloud_files.update(
                        {f"{folder}_{item['Name']}" for item in response["value"]}
                    )
            return cloud_files
        except ConnectionError as error:
            raise SharePointFetchingError(reason=error)

    async def fetch_thies_file_names(self) -> set[str]:
        """Fetch file names from the THIES FTP server."""
        try:
            avg_files = await self.thies_ftp_client.list_files(
                FtpListFilesArgs(path=c.FTP_SERVER_PATH_AVG_FILES)
            )
            ext_files = await self.thies_ftp_client.list_files(
                FtpListFilesArgs(path=c.FTP_SERVER_PATH_EXT_FILES)
            )
            return {f"AVG_{name}" for name in avg_files} | {
                f"EXT_{name}" for name in ext_files
            }
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def fetch_thies_file_content(self) -> dict[str, bytes]:
        """Fetch the content of files from the THIES FTP server."""
        try:
            content_files = {}
            for file in self.uploading:
                origin, filename = file.split("_", 1)
                file_path = (
                    f"{c.FTP_SERVER_PATH_AVG_FILES}/{filename}"
                    if origin == "AVG"
                    else f"{c.FTP_SERVER_PATH_EXT_FILES}/{filename}"
                )
                content = await self.thies_ftp_client.read_file(
                    FtpReadFileArgs(file_path)
                )
                content_files[filename] = content
            return content_files
        except ConnectionRefusedError as error:
            raise ThiesConnectionError(reason=error)
        except ConnectionAbortedError as error:
            raise ThiesFetchingError(reason=error)

    async def execute(self) -> dict:
        """Synchronize data from the THIES Center to the cloud."""
        try:
            thies_files = await self.fetch_thies_file_names()
        except RuntimeError as error:
            raise FtpClientError(error)
        try:
            cloud_files = await self.fetch_cloud_file_names()
        except RuntimeError as error:
            raise SharepointClient(error)
        self.uploading = thies_files - cloud_files
        if not self.uploading:
            raise EmptyDataError(reason="No files to upload.")

        thies_file_contents = await self.fetch_thies_file_content()
        data = parse_execute_response(thies_file_contents)
        return data



from typing import Any

from aiohttp import ClientError, ClientSession

from rcer_iot_client_pkg.libs.async_http_client.async_http_client_contract import (
    AsyncHTTPClientContract,
)
from rcer_iot_client_pkg.libs.async_http_client.types.async_http_client_types import (
    AsyncHttpClientInitArgs,
    GetArgs,
    UploadFileArgs,
)


class AioHttpClient(AsyncHTTPClientContract):
    def __init__(self, args: AsyncHttpClientInitArgs) -> None:
        self.access_token = args.access_token
        self.base_url = args.base_url
        self.headers = self._build_headers()
        self.session: ClientSession | None = None

    def _build_headers(self) -> dict:
        return {"Authorization": f"Bearer {self.access_token}"}

    async def __aenter__(self) -> "AioHttpClient":
        self.session = ClientSession(headers=self.headers, base_url=self.base_url)
        return self

    async def __aexit__(
        self, _exc_type: type[BaseException], _exc_val: BaseException, _exc_tb: Any
    ) -> None:
        await self.session.close()

    async def get(self, args: GetArgs) -> dict[str, Any]:
        try:
            endpoint, params = args.endpoint.lstrip("/"), args.params
            response = await self.session.get(endpoint, params=params)
            response.raise_for_status()
            return await response.json()
        except ClientError as error:
            raise ConnectionError(error) from error

    async def upload_file(self, args: UploadFileArgs) -> dict[str, Any]:
        try:
            endpoint, file_bytes = args.endpoint.lstrip("/"), args.file_bytes
            response = await self.session.put(endpoint, data=file_bytes)
            response.raise_for_status()
            return await response.json()
        except ClientError as error:
            raise ConnectionError(error) from error
